@@ -10,6 +10,7 @@ import com.fasterxml.jackson.databind.JsonNode;
 import com.fasterxml.jackson.databind.util.ArrayBuilders;
 import com.fasterxml.jackson.databind.util.ClassUtil;
 import com.fasterxml.jackson.databind.util.LRUMap;
+import com.fasterxml.jackson.databind.util.LookupCache;
 
 /**
  * Class used for creating concrete {@link JavaType} instances,
@@ -124,7 +125,7 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
      * actual generic types), we will use small cache to avoid repetitive
      * resolution of core types
      */
-    protected final LRUMap<Object,JavaType> _typeCache;
+    protected final LookupCache<Object,JavaType> _typeCache;
 
     /*
     /**********************************************************
@@ -152,15 +153,24 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
      */
 
     private TypeFactory() {
-        this(null);
+        this((LookupCache<Object,JavaType>) null);
     }
 
     /**
      * @since 2.8
+     * @deprecated Since 2.12
      */
+    @Deprecated // since 2.12
     protected TypeFactory(LRUMap<Object,JavaType> typeCache) {
+        this((LookupCache<Object,JavaType>) typeCache);
+    }
+
+    /**
+     * @since 2.12
+     */
+    protected TypeFactory(LookupCache<Object,JavaType> typeCache) {
         if (typeCache == null) {
-            typeCache = new LRUMap<Object,JavaType>(16, 200);
+            typeCache = new LRUMap<>(16, 200);
         }
         _typeCache = typeCache;
         _parser = new TypeParser(this);
@@ -168,11 +178,24 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
         _classLoader = null;
     }
 
+    /**
+     * @deprecated Since 2.12
+     */
+    @Deprecated // since 2.12
     protected TypeFactory(LRUMap<Object,JavaType> typeCache, TypeParser p,
             TypeModifier[] mods, ClassLoader classLoader)
+    {
+        this((LookupCache<Object,JavaType>) typeCache, p, mods, classLoader);
+    }
+
+    /**
+     * @since 2.12
+     */
+    protected TypeFactory(LookupCache<Object,JavaType> typeCache, TypeParser p,
+                          TypeModifier[] mods, ClassLoader classLoader)
     {
         if (typeCache == null) {
-            typeCache = new LRUMap<Object,JavaType>(16, 200);
+            typeCache = new LRUMap<>(16, 200);
         }
         _typeCache = typeCache;
         // As per [databind#894] must ensure we have back-linkage from TypeFactory:
@@ -188,7 +211,7 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
      */
     public TypeFactory withModifier(TypeModifier mod) 
     {
-        LRUMap<Object,JavaType> typeCache = _typeCache;
+        LookupCache<Object,JavaType> typeCache = _typeCache;
         TypeModifier[] mods;
         if (mod == null) { // mostly for unit tests
             mods = null;
@@ -221,11 +244,24 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
      * bigger maximum size.
      *
      * @since 2.8
+     * @deprecated Since 2.12
      */
+    @Deprecated // since 2.12
     public TypeFactory withCache(LRUMap<Object,JavaType> cache)  {
         return new TypeFactory(cache, _parser, _modifiers, _classLoader);
     }
 
+    /**
+     * Mutant factory method that will construct new {@link TypeFactory} with
+     * identical settings except for different cache; most likely one with
+     * bigger maximum size.
+     *
+     * @since 2.12
+     */
+    public TypeFactory withCache(LookupCache<Object,JavaType> cache)  {
+        return new TypeFactory(cache, _parser, _modifiers, _classLoader);
+    }
+
     /**
      * Method used to access the globally shared instance, which has
      * no custom configuration. Used by <code>ObjectMapper</code> to
@@ -240,7 +276,7 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
      * if you know there is a problem with retention of type definitions;
      * the most likely (and currently only known) problem is retention
      * of {@link Class} instances via {@link JavaType} reference.
-     * 
+     *
      * @since 2.4.1
      */
     public void clearCache() {
@@ -687,10 +723,10 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
         }
         return type1;
     }
-    
+
     /*
     /**********************************************************
-    /* Public factory methods
+    /* Public general-purpose factory methods
     /**********************************************************
      */
 
@@ -698,10 +734,6 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
         return _fromAny(null, type, EMPTY_BINDINGS);
     }
 
-    public JavaType constructType(Type type, TypeBindings bindings) {
-        return _fromAny(null, type, bindings);
-    }
-    
     public JavaType constructType(TypeReference<?> typeRef)
     {
         // 19-Oct-2015, tatu: Simpler variant like so should work
@@ -726,6 +758,53 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
         */
     }
 
+    /**
+     * Method to call when resolving types of {@link java.lang.reflect.Member}s
+     * like Fields, Methods and Constructor parameters and there is a
+     * {@link TypeBindings} (that describes binding of type parameters within
+     * context) to pass.
+     * This is typically used only by code in databind itself.
+     * 
+     * @param type Type of a {@link java.lang.reflect.Member} to resolve
+     * @param contextBindings Type bindings from the context, often class in which
+     *     member declared but may be subtype of that type (to bind actual bound
+     *     type parametrers). Not used if {@code type} is of type {@code Class<?>}.
+     *
+     * @return Fully resolve type
+     *
+     * @since 2.12 as replacement for deprecated {@link #constructType(Type, TypeBindings)} 
+     */
+    public JavaType resolveMemberType(Type type, TypeBindings contextBindings) {
+        return _fromAny(null, type, contextBindings);
+    }
+
+    /*
+    /**********************************************************
+    /* Deprecated public factory methods
+    /**********************************************************
+     */
+
+    /**
+     * Method that you very likely should NOT be using -- you need to know a lot
+     * about internal details of {@link TypeBindings} and even then it will probably
+     * not do what you want.
+     * Usually you would instead want to call one of {@code constructXxxType()}
+     * methods (where {@code Xxx} would be "Array", "Collection[Like]", "Map[Like]"
+     * or "Parametric").
+     *
+     * @deprecated Since 2.12
+     */
+    @Deprecated // since 2.12
+    public JavaType constructType(Type type, TypeBindings bindings) {
+        // 15-Jun-2020, tatu: To resolve (parts of) [databind#2796], need to
+        //    call _fromClass() directly if we get `Class` argument
+        if (type instanceof Class<?>) {
+            JavaType resultType = _fromClass(null, (Class<?>) type, bindings);
+            return _applyModifiers(type, resultType);
+        }
+        return _fromAny(null, type, bindings);
+    }
+
     /**
      * @deprecated Since 2.7 (accidentally removed in 2.7.0; added back in 2.7.1)
      */
@@ -874,8 +953,7 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
     /**
      * Method for constructing a {@link MapType} instance
      *<p>
-     * NOTE: type modifiers are NOT called on constructed type itself; but are called
-     * for contained types.
+     * NOTE: type modifiers are NOT called on constructed type itself.
      */
     public MapType constructMapType(Class<? extends Map> mapClass, JavaType keyType, JavaType valueType) {
         TypeBindings bindings = TypeBindings.createIfNeeded(mapClass, new JavaType[] { keyType, valueType });
@@ -915,8 +993,7 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
     /**
      * Method for constructing a {@link MapLikeType} instance
      *<p>
-     * NOTE: type modifiers are NOT called on constructed type itself; but are called
-     * for contained types.
+     * NOTE: type modifiers are NOT called on constructed type itself.
      */
     public MapLikeType constructMapLikeType(Class<?> mapClass, JavaType keyType, JavaType valueType) {
         // 19-Oct-2015, tatu: Allow case of no-type-variables, since it seems likely to be
@@ -932,7 +1009,7 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
     /**
      * Method for constructing a type instance with specified parameterization.
      *<p>
-     * NOTE: was briefly deprecated for 2.6.
+     * NOTE: type modifiers are NOT called on constructed type itself.
      */
     public JavaType constructSimpleType(Class<?> rawType, JavaType[] parameterTypes) {
         return _fromClass(null, rawType, TypeBindings.create(rawType, parameterTypes));
@@ -953,11 +1030,17 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
     } 
 
     /**
+     * Method for constructing a {@link ReferenceType} instance with given type parameter
+     * (type MUST take one and only one type parameter)
+     *<p>
+     * NOTE: type modifiers are NOT called on constructed type itself.
+     *
      * @since 2.6
      */
     public JavaType constructReferenceType(Class<?> rawType, JavaType referredType)
     {
-        return ReferenceType.construct(rawType, null, // no bindings
+        return ReferenceType.construct(rawType,
+                TypeBindings.create(rawType, referredType), // [databind#2091]
                 null, null, // or super-class, interfaces?
                 referredType);
     }
@@ -985,23 +1068,13 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
      * type {@code List<Set<Integer>>}, you could
      * call
      *<pre>
-     *  JavaType inner = TypeFactory.constructParametricType(Set.class, Set.class, Integer.class);
-     *  return TypeFactory.constructParametricType(ArrayList.class, List.class, inner);
+     *  JavaType inner = TypeFactory.constructParametricType(Set.class, Integer.class);
+     *  return TypeFactory.constructParametricType(List.class, inner);
      *</pre>
      *<p>
-     * The reason for first two arguments to be separate is that parameterization may
-     * apply to a super-type. For example, if generic type was instead to be
-     * constructed for {@code ArrayList<Integer>}, the usual call would be:
-     *<pre>
-     *  TypeFactory.constructParametricType(ArrayList.class, List.class, Integer.class);
-     *</pre>
-     * since parameterization is applied to {@link java.util.List}.
-     * In most cases distinction does not matter, but there are types where it does;
-     * one such example is parameterization of types that implement {@link java.util.Iterator}.
-     *<p>
-     * NOTE: type modifiers are NOT called on constructed type.
+     * NOTE: since 2.11.2 {@link TypeModifier}s ARE called on result (fix for [databind#2796])
      * 
-     * @param parametrized Actual full type
+     * @param parametrized Type-erased type to parameterize
      * @param parameterClasses Type parameters to apply
      *
      * @since 2.5 NOTE: was briefly deprecated for 2.6
@@ -1020,30 +1093,45 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
      * represents a parameterized type. For example, to represent
      * type {@code List<Set<Integer>>}, you could
      *<pre>
-     *  JavaType inner = TypeFactory.constructParametricType(Set.class, Set.class, Integer.class);
-     *  return TypeFactory.constructParametricType(ArrayList.class, List.class, inner);
+     *  JavaType inner = TypeFactory.constructParametricType(Set.class, Integer.class);
+     *  return TypeFactory.constructParametricType(List.class, inner);
      *</pre>
      *<p>
-     * The reason for first two arguments to be separate is that parameterization may
-     * apply to a super-type. For example, if generic type was instead to be
-     * constructed for {@code ArrayList<Integer>}, the usual call would be:
-     *<pre>
-     *  TypeFactory.constructParametricType(ArrayList.class, List.class, Integer.class);
-     *</pre>
-     * since parameterization is applied to {@link java.util.List}.
-     * In most cases distinction does not matter, but there are types where it does;
-     * one such example is parameterization of types that implement {@link java.util.Iterator}.
-     *<p>
-     * NOTE: type modifiers are NOT called on constructed type.
+     * NOTE: since 2.11.2 {@link TypeModifier}s ARE called on result (fix for [databind#2796])
      *
      * @param rawType Actual type-erased type
      * @param parameterTypes Type parameters to apply
      *
-     * @since 2.5 NOTE: was briefly deprecated for 2.6
+     * @return Fully resolved type for given base type and type parameters
      */
     public JavaType constructParametricType(Class<?> rawType, JavaType... parameterTypes)
     {
-        return _fromClass(null, rawType, TypeBindings.create(rawType, parameterTypes));
+        return constructParametricType(rawType, TypeBindings.create(rawType, parameterTypes));
+    }
+
+    /**
+     * Factory method for constructing {@link JavaType} that
+     * represents a parameterized type. The type's parameters are
+     * specified as an instance of {@link TypeBindings}. This
+     * is useful if you already have the type's parameters such
+     * as those found on {@link JavaType}. For example, you could call
+     * <pre>
+     *   return TypeFactory.constructParametricType(ArrayList.class, javaType.getBindings());
+     * </pre>
+     * This effectively applies the parameterized types from one
+     * {@link JavaType} to another class.
+     *
+     * @param rawType Actual type-erased type
+     * @param parameterTypes Type bindings for the raw type
+     *
+     * @since 2.12
+     */
+    public JavaType constructParametricType(Class<?> rawType, TypeBindings parameterTypes)
+    {
+        // 16-Jul-2020, tatu: Since we do not call `_fromAny()`, need to make
+        //   sure `TypeModifier`s are applied:
+        JavaType resultType =  _fromClass(null, rawType, parameterTypes);
+        return _applyModifiers(rawType, resultType);
     }
 
     /**
@@ -1154,7 +1242,8 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
         } else {
             List<JavaType> typeParams = bindings.getTypeParameters();
             // ok to have no types ("raw")
-            switch (typeParams.size()) {
+            final int pc = typeParams.size();
+            switch (pc) {
             case 0: // acceptable?
                 kt = vt = _unknownType();
                 break;
@@ -1163,7 +1252,9 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
                 vt = typeParams.get(1);
                 break;
             default:
-                throw new IllegalArgumentException("Strange Map type "+rawClass.getName()+": cannot determine type parameters");
+                throw new IllegalArgumentException(String.format(
+"Strange Map type %s with %d type parameter%s (%s), can not resolve",
+ClassUtil.nameOf(rawClass), pc, (pc == 1) ? "" : "s", bindings));
             }
         }
         return MapType.construct(rawClass, bindings, superClass, superInterfaces, kt, vt);
@@ -1274,51 +1365,58 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
      * as Java typing returned from <code>getGenericXxx</code> methods
      * (usually for a return or argument type).
      */
-    protected JavaType _fromAny(ClassStack context, Type type, TypeBindings bindings)
+    protected JavaType _fromAny(ClassStack context, Type srcType, TypeBindings bindings)
     {
         JavaType resultType;
 
         // simple class?
-        if (type instanceof Class<?>) {
+        if (srcType instanceof Class<?>) {
             // Important: remove possible bindings since this is type-erased thingy
-            resultType = _fromClass(context, (Class<?>) type, EMPTY_BINDINGS);
+            resultType = _fromClass(context, (Class<?>) srcType, EMPTY_BINDINGS);
         }
         // But if not, need to start resolving.
-        else if (type instanceof ParameterizedType) {
-            resultType = _fromParamType(context, (ParameterizedType) type, bindings);
+        else if (srcType instanceof ParameterizedType) {
+            resultType = _fromParamType(context, (ParameterizedType) srcType, bindings);
         }
-        else if (type instanceof JavaType) { // [databind#116]
+        else if (srcType instanceof JavaType) { // [databind#116]
             // no need to modify further if we already had JavaType
-            return (JavaType) type;
+            return (JavaType) srcType;
         }
-        else if (type instanceof GenericArrayType) {
-            resultType = _fromArrayType(context, (GenericArrayType) type, bindings);
+        else if (srcType instanceof GenericArrayType) {
+            resultType = _fromArrayType(context, (GenericArrayType) srcType, bindings);
         }
-        else if (type instanceof TypeVariable<?>) {
-            resultType = _fromVariable(context, (TypeVariable<?>) type, bindings);
+        else if (srcType instanceof TypeVariable<?>) {
+            resultType = _fromVariable(context, (TypeVariable<?>) srcType, bindings);
         }
-        else if (type instanceof WildcardType) {
-            resultType = _fromWildcard(context, (WildcardType) type, bindings);
+        else if (srcType instanceof WildcardType) {
+            resultType = _fromWildcard(context, (WildcardType) srcType, bindings);
         } else {
             // sanity check
-            throw new IllegalArgumentException("Unrecognized Type: "+((type == null) ? "[null]" : type.toString()));
+            throw new IllegalArgumentException("Unrecognized Type: "+((srcType == null) ? "[null]" : srcType.toString()));
         }
         // 21-Feb-2016, nateB/tatu: as per [databind#1129] (applied for 2.7.2),
         //   we do need to let all kinds of types to be refined, esp. for Scala module.
-        if (_modifiers != null) {
-            TypeBindings b = resultType.getBindings();
-            if (b == null) {
-                b = EMPTY_BINDINGS;
-            }
-            for (TypeModifier mod : _modifiers) {
-                JavaType t = mod.modifyType(resultType, type, b, this);
-                if (t == null) {
-                    throw new IllegalStateException(String.format(
-                            "TypeModifier %s (of type %s) return null for type %s",
-                            mod, mod.getClass().getName(), resultType));
-                }
-                resultType = t;
+        return _applyModifiers(srcType, resultType);
+    }
+
+    protected JavaType _applyModifiers(Type srcType, JavaType resolvedType)
+    {
+        if (_modifiers == null) {
+            return resolvedType;
+        }
+        JavaType resultType = resolvedType;
+        TypeBindings b = resultType.getBindings();
+        if (b == null) {
+            b = EMPTY_BINDINGS;
+        }
+        for (TypeModifier mod : _modifiers) {
+            JavaType t = mod.modifyType(resultType, srcType, b, this);
+            if (t == null) {
+                throw new IllegalStateException(String.format(
+                        "TypeModifier %s (of type %s) return null for type %s",
+                        mod, mod.getClass().getName(), resultType));
             }
+            resultType = t;
         }
         return resultType;
     }
@@ -1446,7 +1544,6 @@ public class TypeFactory // note: was final in 2.9, removed from 2.10
         if (bindings == null) {
             bindings = EMPTY_BINDINGS;
         }
-        
         // Quite simple when we resolving exact class/interface; start with that
         if (rawType == Map.class) {
             return _mapType(rawType, bindings, superClass, superInterfaces);
